use std::{
    convert::Infallible,
    error::Error,
    fmt::Display,
    num::ParseIntError,
    ops::Deref,
    path::{Path, PathBuf},
    process::Command,
    str::FromStr,
};

use crate::fs::{
    id_mapping::{IdMapping, ParseIdRangeError},
    mount::MountOptions,
    AsIter, FileSystem, Mountpoint,
};
use crate::macros::*;

use super::Stack;

macro_rules! impl_wrapper {
    ($name:ident($inner:ty), $err:ty, $from_str:block) => {
        impl Deref for $name {
            type Target = $inner;

            fn deref(&self) -> &Self::Target {
                &self.0
            }
        }

        impl FromStr for $name {
            type Err = $err;

            fn from_str(s: &str) -> Result<Self, Self::Err> {
                $from_str
                //Ok(Self($target::from_str(s)?))
            }
        }

        impl Display for $name {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(f, "{}", self.0)
            }
        }
    };
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct Test(usize);

impl_wrapper!(Test(usize), ParseIntError, {
    Ok(Test(usize::from_str(s)?))
});

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseMaxIdleThreadsError(ParseIntError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseMaxThreadsError(ParseIntError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSquashToIdError(ParseIntError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSquashToUidError(ParseSquashToIdError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSquashToGidError(ParseSquashToIdError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseUidMappingError(ParseIdRangeError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseGidMappingError(ParseIdRangeError);

impl_wrapper_err_for!(
    ParseMaxIdleThreadsError(ParseIntError),
    "invalid max idle threads"
);

impl_wrapper_err_for!(ParseMaxThreadsError(ParseIntError), "invalid max threads");
impl_wrapper_err_for!(ParseSquashToIdError(ParseIntError), "invalid id");
impl_wrapper_err_for!(ParseSquashToUidError(ParseSquashToIdError), "invalid uid");
impl_wrapper_err_for!(ParseSquashToGidError(ParseSquashToIdError), "invalid gid");
impl_wrapper_err_for!(ParseUidMappingError(ParseIdRangeError), "invalid uid");
impl_wrapper_err_for!(ParseGidMappingError(ParseIdRangeError), "invalid gid");

#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct LowerDirs(Vec<PathBuf>);

impl Deref for LowerDirs {
    type Target = Vec<PathBuf>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl Display for LowerDirs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            self.0
                .iter()
                .map(|v| v.to_string_lossy())
                .fold(String::new(), |a, b| a + &b)
        )
    }
}

impl FromStr for LowerDirs {
    type Err = Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(s.split(',').map(Into::into).collect()))
    }
}

impl From<Vec<&str>> for LowerDirs {
    fn from(value: Vec<&str>) -> Self {
        Self(value.into_iter().map(PathBuf::from).collect())
    }
}

impl From<Vec<String>> for LowerDirs {
    fn from(value: Vec<String>) -> Self {
        Self(value.into_iter().map(PathBuf::from).collect())
    }
}

impl From<Vec<&Path>> for LowerDirs {
    fn from(value: Vec<&Path>) -> Self {
        Self(value.into_iter().map(Path::to_path_buf).collect())
    }
}

impl From<Vec<PathBuf>> for LowerDirs {
    fn from(value: Vec<PathBuf>) -> Self {
        Self(value)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct MaxIdleThreads(Option<usize>);

impl Default for MaxIdleThreads {
    fn default() -> Self {
        Self(None)
    }
}

impl FromStr for MaxIdleThreads {
    type Err = ParseMaxIdleThreadsError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "-1" => Ok(Self(None)),
            s => Ok(usize::from_str_radix(s, 10).map(Some).map(Self)?),
        }
    }
}

impl Display for MaxIdleThreads {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.0 {
            Some(n) => write!(f, "{n}"),
            None => write!(f, "-1"),
        }
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct MaxThreads(usize);

impl Default for MaxThreads {
    fn default() -> Self {
        Self(10)
    }
}

impl FromStr for MaxThreads {
    type Err = ParseMaxThreadsError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(usize::from_str_radix(s, 10).map(Self)?)
    }
}

impl Display for MaxThreads {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct SquashToId(usize);

impl FromStr for SquashToId {
    type Err = ParseSquashToIdError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(usize::from_str_radix(s, 10).map(Self)?)
    }
}

impl Display for SquashToId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct SquashToUid(SquashToId);

impl Deref for SquashToUid {
    type Target = SquashToId;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl FromStr for SquashToUid {
    type Err = ParseSquashToUidError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(SquashToId::from_str(s)?))
    }
}

impl Display for SquashToUid {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct SquashToGid(SquashToId);

impl Deref for SquashToGid {
    type Target = SquashToId;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl FromStr for SquashToGid {
    type Err = ParseSquashToGidError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(SquashToId::from_str(s)?))
    }
}

impl Display for SquashToGid {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}
#[derive(Clone, Debug, PartialEq, Eq)]
pub struct GidMapping(IdMapping);

impl Deref for GidMapping {
    type Target = IdMapping;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl FromStr for GidMapping {
    type Err = ParseGidMappingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(IdMapping::from_str(s)?))
    }
}

impl Display for GidMapping {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct GidMapping(IdMapping);

impl Deref for GidMapping {
    type Target = IdMapping;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl FromStr for GidMapping {
    type Err = ParseGidMappingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(IdMapping::from_str(s)?))
    }
}

impl Display for GidMapping {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ParseFuseOverlayOptionError {
    MaxIdleThreads(ParseMaxIdleThreadsError),
    MaxThreads(ParseMaxThreadsError),
    SquashToUid(ParseSquashToUidError),
    SquashToGid(ParseSquashToGidError),
    UidMapping(ParseUidMappingError),
    GidMapping(ParseGidMappingError),
}

impl Display for ParseFuseOverlayOptionError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::MaxIdleThreads(e) => write!(f, "{e}"),
            Self::MaxThreads(e) => write!(f, "{e}"),
            Self::SquashToUid(e) => write!(f, "{e}"),
            Self::SquashToGid(e) => write!(f, "{e}"),
            Self::UidMapping(e) => write!(f, "{e}"),
            Self::GidMapping(e) => write!(f, "{e}"),
        }
    }
}

impl Error for ParseFuseOverlayOptionError {}

impl_from_variants!(
    ParseFuseOverlayOptionError,
    MaxIdleThreads(ParseMaxIdleThreadsError),
    MaxThreads(ParseMaxThreadsError)
);

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum FuseOverlayOption {
    LowerDir(LowerDirs),
    UpperDir(PathBuf),
    WorkDir(PathBuf),
    CloneFd,
    MaxIdleThreads(MaxIdleThreads),
    MaxThreads(MaxThreads),
    AllowOther,
    AllowRoot,
    SquashToRoot,
    SquashToUid(SquashToUid),
    SquashToGid(SquashToGid),
    StaticNLink,
    NoAcl,
    UidMapping(IdMapping),
    GidMapping(IdMapping),
}

impl_from_variants!(
    FuseOverlayOption,
    LowerDir(LowerDirs),
    MaxIdleThreads(MaxIdleThreads),
    MaxThreads(MaxThreads),
    SquashToUid(SquashToUid),
    SquashToGid(SquashToGid)
);

impl FromStr for FuseOverlayOption {
    type Err = ParseFuseOverlayOptionError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let s = s.trim().trim_start_matches('-');

        let option = match s.split_once(|delim| delim == ' ' || delim == '=') {
            Some((option, value)) => (option, Some(value)),
            None => (s, None),
        };

        match option {
            ("lowerdir", Some(args)) => Ok(LowerDirs::from_str(args).unwrap().into()),
            ("upperdir", Some(args)) => Ok(Self::UpperDir(PathBuf::from(args))),
            ("workdir", Some(args)) => Ok(Self::WorkDir(PathBuf::from(args))),
            ("clonefd", None) => Ok(Self::CloneFd),
            ("max_idle_threads", Some(args)) => Ok(MaxIdleThreads::from_str(args).into()?),
            ("max_threads", Some(args)) => Ok(MaxThreads::from_str(args).into()?),
            ("allow_other", None) => Ok(Self::AllowOther),
            ("allow_root", None) => Ok(Self::AllowRoot),
            ("squash_to_root", None) => Ok(Self::SquashToRoot),
            ("squash_to_uid", Some(uid)) => Ok(SquashToUid::from_str(uid).into()?),
            ("squash_to_gid", Some(gid)) => Ok(SquashToGid::from_str(gid).into()?),
            ("static_nlink", None) => Ok(Self::StaticNLink),
            ("noacl", None) => Ok(Self::NoAcl),
            ("uidmapping", Some(ids)) => Ok(Self::UidMapping(IdMapping::from_str(ids)?)),
            ("gidmapping", Some(ids)) => Ok(Self::GidMapping(IdMapping::from_str(ids)?)),
        }
    }
}

impl Display for FuseOverlayOption {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "-o ")?;

        match self {
            FuseOverlayOption::LowerDir(lower) => write!(f, "{lower}"),
            FuseOverlayOption::UpperDir(path) => write!(f, "{}", path.to_string_lossy()),
            FuseOverlayOption::WorkDir(path) => write!(f, "{}", path.to_string_lossy()),
            FuseOverlayOption::CloneFd => write!(f, "clone_fd"),
            FuseOverlayOption::MaxIdleThreads(n) => write!(f, "max_idle_threads={n}"),
            FuseOverlayOption::MaxThreads(n) => write!(f, "max_threads={n}"),
            FuseOverlayOption::AllowOther => write!(f, "allow_other"),
            FuseOverlayOption::AllowRoot => write!(f, "allow_root"),
            FuseOverlayOption::SquashToRoot => write!(f, "squash_to_root"),
            FuseOverlayOption::SquashToUid(uid) => write!(f, "squash_to_uid={uid}"),
            FuseOverlayOption::SquashToGid(gid) => write!(f, "squash_to_gid={gid}"),
            FuseOverlayOption::StaticNLink => write!(f, "static_nlink"),
            FuseOverlayOption::NoAcl => write!(f, "noacl"),
            FuseOverlayOption::UidMapping(map) => write!(f, "uidmapping={map}"),
            FuseOverlayOption::GidMapping(map) => write!(f, "gidmapping={map}"),
        }
    }
}

#[derive(Debug)]
pub struct FuseOverlay {
    lower_dir: usize,
    upper_dir: Option<usize>,
    work_dir: Option<usize>,
    options: MountOptions<FuseOverlayOption>,
}

impl FuseOverlay {
    pub fn new(lowerdir: impl Into<LowerDirs>) -> Self {
        Self {
            lower_dir: 0,
            upper_dir: None,
            work_dir: None,
            options: MountOptions(vec![FuseOverlayOption::LowerDir(lowerdir.into())]),
        }
    }
}

impl Stack for FuseOverlay {
    fn lower_dirs(&self) -> impl Iterator<Item = &std::path::Path> {
        match self.options.get(self.lower_dir) {
            Some(FuseOverlayOption::LowerDir(lower)) => lower.as_iter().map(AsRef::as_ref),
            _ => panic!("invalid lowerdir option"),
        }
    }

    fn upper_dir(&self) -> Option<&std::path::Path> {
        let upper = self.upper_dir.and_then(|i| self.options.get(i));

        match upper {
            Some(FuseOverlayOption::UpperDir(upper)) => Some(upper),
            _ => None,
        }
    }

    fn work_dir(&self) -> Option<&std::path::Path> {
        let work = self.work_dir.and_then(|i| self.options.get(i));

        match work {
            Some(FuseOverlayOption::WorkDir(work)) => Some(work),
            _ => None,
        }
    }
}

impl Mountpoint for FuseOverlay {
    fn options(&self) -> impl Iterator<Item = impl Display> {
        self.options.as_iter()
    }
}

impl FileSystem for FuseOverlay {
    fn mount(&mut self) -> std::io::Result<()> {
        let mut cmd = Command::new("fuse-overlayfs").arg(self.options.to_string());

        if self.fs.lower.len() > 0 {
            cmd.arg(format!("-o lowerdir={}", self.fs.lower));
        };

        if let Some(upper) = &self.fs.upper {
            cmd.arg(format!("-o upperdir={}", upper.to_string_lossy()));
        };

        if let Some(work) = &self.fs.work {
            cmd.arg(format!("-o workdir={}", work.to_string_lossy()));
        };

        cmd.output()?;

        Ok(())
    }

    fn unmount(&mut self) -> std::io::Result<()> {
        if let Some(upper) = &self.fs.upper {
            Command::new("fusermount").arg("-u").arg(upper).output()?;
        }

        Ok(())
    }
}

impl Display for FuseOverlay {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "fuse-overlayfs")
    }
}
